#include <stdio.h>

#include "brd.h"
#include "f4_gpio.h"

#include "s3.h"


void
f4_delay(int ms)
{
  unsigned int n= HAL_GetTick();

  while ((HAL_GetTick() - n) < ms)
    ;
}

/* Init GPIO for bus lines */
void
s3_init(void)
{
  /* Connected, but unused pins */
  f4_gpio_input("PD13");
  f4_gpio_input("PD11");
  f4_gpio_input("PE15");
  f4_gpio_input("PB10");
  f4_gpio_input("PB1");
  f4_gpio_input("PB0");
  f4_gpio_input("PC5");
  f4_gpio_input("PC4");
  f4_gpio_input("PA7");
  f4_gpio_input("PA6");
  f4_gpio_input("PA5");
  f4_gpio_input("PA4");

  /* Control lines */
  f4_gpio_output(S_RST);
  f4_gpio_set(S_RST, 0);
  f4_gpio_output(S_WR);
  f4_gpio_set(S_WR, 0);
  f4_gpio_output(S_RD);
  f4_gpio_set(S_RD, 0);

  /* Data lines */
  f4_gpio_input(S_D0);
  f4_gpio_input(S_D1);
  f4_gpio_input(S_D2);
  f4_gpio_input(S_D3);
  f4_gpio_input(S_D4);
  f4_gpio_input(S_D5);
  f4_gpio_input(S_D6);
  f4_gpio_input(S_D7);

  /* Address lines */
  f4_gpio_output(S_A0);
  f4_gpio_output(S_A1);
  f4_gpio_output(S_A2);
  f4_gpio_output(S_A3);
  f4_gpio_output(S_A4);
  f4_gpio_output(S_A5);
  f4_gpio_output(S_A6);
  f4_gpio_output(S_A7);

  /* Send a RESET */
  f4_gpio_set(S_RST, 1);
  f4_delay(1);
  f4_gpio_set(S_RST, 0);
}

/* Turn ON data bus */
void
s3_d_on()
{
  f4_gpio_output(S_D0);
  f4_gpio_output(S_D1);
  f4_gpio_output(S_D2);
  f4_gpio_output(S_D3);
  f4_gpio_output(S_D4);
  f4_gpio_output(S_D5);
  f4_gpio_output(S_D6);
  f4_gpio_output(S_D7);
}

/* Turn OFF data bus */
void
s3_d_off()
{
  f4_gpio_input(S_D0);
  f4_gpio_input(S_D1);
  f4_gpio_input(S_D2);
  f4_gpio_input(S_D3);
  f4_gpio_input(S_D4);
  f4_gpio_input(S_D5);
  f4_gpio_input(S_D6);
  f4_gpio_input(S_D7);
}

/* Set value of address lines */
void
s3_set_a(int address)
{
  f4_gpio_set(S_A0, address & 0x01);
  f4_gpio_set(S_A1, address & 0x02);
  f4_gpio_set(S_A2, address & 0x04);
  f4_gpio_set(S_A3, address & 0x08);
  f4_gpio_set(S_A4, address & 0x10);
  f4_gpio_set(S_A5, address & 0x20);
  f4_gpio_set(S_A6, address & 0x40);
  f4_gpio_set(S_A7, address & 0x80);
}

/* Set value of data lines */
void
s3_set_d(int value)
{
  f4_gpio_set(S_D0, value & 0x01);
  f4_gpio_set(S_D1, value & 0x02);
  f4_gpio_set(S_D2, value & 0x04);
  f4_gpio_set(S_D3, value & 0x08);
  f4_gpio_set(S_D4, value & 0x10);
  f4_gpio_set(S_D5, value & 0x20);
  f4_gpio_set(S_D6, value & 0x40);
  f4_gpio_set(S_D7, value & 0x80);
}

/* Read value of data lines */
int
s3_get_d()
{
  int r= 0;

  r|= f4_gpio_get(S_D0) << 0;
  r|= f4_gpio_get(S_D1) << 1;
  r|= f4_gpio_get(S_D2) << 2;
  r|= f4_gpio_get(S_D3) << 3;
  r|= f4_gpio_get(S_D4) << 4;
  r|= f4_gpio_get(S_D5) << 5;
  r|= f4_gpio_get(S_D6) << 6;
  r|= f4_gpio_get(S_D7) << 7;

  return r;
}

/* Set RD line */
void
s3_set_rd(int value)
{
  f4_gpio_set(S_RD, value);
}

/* Set WR line */
void
s3_set_wr(int value)
{
  f4_gpio_set(S_WR, value);
}

/* Set RST line */
void
s3_set_rst(int value)
{
  f4_gpio_set(S_RST, value);
}

/* Read value of addressed register */
int
s3_read(int address)
{
  int r;
  s3_d_off();           /* Setup data lines for READ */
  s3_set_a(address);    /* Set address lines */
  f4_delay(1);          /* Data setup time */
  s3_set_rd(1);         /* Pulse on RD */
  f4_delay(1);          /* Data hold time */
  r= s3_get_d();        /* Read value from data lines */
  s3_set_rd(0);         /* Release RD line */
  return r;             /* Return read value */
}

/* Write value to addressed register */
int
s3_write(int address, int value)
{
  s3_d_on();            /* Setup data lines for WRITE */
  s3_set_a(address);    /* Set address lines */
  s3_set_d(value);      /* Put value on data lines */
  f4_delay(1);          /* Data setup time */
  s3_set_wr(1);         /* Pulse on WR line */
  f4_delay(1);          /* Data hold time */
  s3_set_wr(0);         /* Release WR line */
  f4_delay(1);          /* Some more delay... */
  s3_d_off();           /* Release data lines */
  return value;         /* Return written value */
}

void
s3_printd(int v)
{
	char s[5];
	int i;
	snprintf(s, 5, "%-d", v);
	for (i= 0; (i<4) && (s[i]); i++)
	{
		if (s[i]==' ')
		{
			s3_write(11-i, 0);
		}
		else
		{
			s3_write(3-i, s[i]-'0');
		}
	}
	for ( ; i<4; i++)
		s3_write(11-i, 0);
}

/* End of s3.c */
